from __future__ import annotations

from typing import Any, Union, Optional, TypedDict, NotRequired, cast
from collections.abc import Mapping

from datetime import timedelta
from pathlib import Path
from basyx.aas import model

from .utils import timedelta_to_relativedelta

ListValueType = list[Optional['ElementValueType']]
CollectionValueType = dict[str, Optional['ElementValueType']]
ElementValueType = Union[model.ValueDataType, 'FileValue', 'RangeValue', 'MultiLanguagePropertyValue',
                        ListValueType, CollectionValueType]

PropertyJsonValueType = Union[str,int,float,bool]
ListJsonValueType = list['ElementJsonValueType']
CollectionJsonValueType = dict[str, 'ElementJsonValueType']
ElementJsonValueType = Union[PropertyJsonValueType, 'FileJsonValue', 'RangeValue', 'MultiLanguagePropertyValue',
                            ListJsonValueType, CollectionJsonValueType]

ElementValueDict = dict[str, Optional[ElementValueType]]


def update_value_dict(target:ElementValueDict,
                      new_values:Mapping[str, ElementValueType|'ElementReference'|model.SubmodelElement|None]) -> None:
  from .reference import ElementReference
  for key, new_value in new_values.items():
    if key not in target:
      continue
    if new_value is None:
      target[key] = None
    elif isinstance(new_value, ElementReference):
      target[key] = new_value.read_value()
    elif isinstance(new_value, model.SubmodelElement):
      target[key] = get_value(new_value)
    else:
      target[key] = new_value

def to_file_value(path:str, content_type:Optional[str]=None) -> FileValue:
  if content_type is None:
    from tika import parser
    parsed = parser.from_file(path)
    mime_type = parsed["metadata"]["Content-Type"] # type: ignore
  else:
    mime_type = content_type
  return { 'content_type': mime_type, 'value': Path(path).name }

class FileValue(TypedDict):
  content_type: str
  value: Optional[str]
class FileJsonValue(TypedDict):
  contentType: str
  value: Optional[str]

class RangeValue(TypedDict):
  min: Optional[model.ValueDataType]
  max: Optional[model.ValueDataType]

MultiLanguagePropertyValue = dict[str,str]
MultiLanguagePropertyJsonValue = list[dict[str,str]]


def get_value(sme:model.SubmodelElement) -> Optional[ElementValueType]:
  """SubmodelElement 객체의 값을 반환한다.

  Args:
      sme (model.SubmodelElement): 값을 얻어올 SubmodelElement 객체.

  Returns:
      Optional[ElementValueType]: SubmodelElement 객체의 값.
  """
  assert sme is not None
  match sme:
    case model.Property():
      return sme.value
    case model.SubmodelElementCollection():
      return { str(member.id_short):get_value(member) for member in sme.value }
    case model.SubmodelElementList():
      return [ get_value(member) for member in sme.value ]
    case model.File():
      return { 'content_type': sme.content_type, 'value': sme.value }
    case model.Range():
      return { 'min': sme.min, 'max': sme.max }
    case model.MultiLanguageProperty():
      if sme.value is None:
        return None
      return { lang:text for lang, text in sme.value.items() }
    case _:
      raise NotImplementedError(f"Unknown SubmodelElement type: {type(sme)}")


def update_element_with_value(sme:model.SubmodelElement, value:Optional[ElementValueType]) -> None:
  """SubmodelElement 객체의 값을 변경한다.

  Args:
      sme (model.SubmodelElement): 값을 변경할 SubmodelElement 객체.
      value (Optional[ElementValueType]): 변경할 값.
  """
  if value is None:
    return
  match sme:
    case model.Property():
      value = timedelta_to_relativedelta(value) if value is not None and isinstance(value, timedelta) else value
      sme.value = value
    case model.SubmodelElementCollection():
      assert isinstance(value, dict)
      for member in sme.value:
        member_value = value.get(str(member.id_short))
        if member_value is not None:
          update_element_with_value(member, member_value)
    case model.SubmodelElementList():
      assert isinstance(value, list)
      for member, member_value in zip[tuple[model.SubmodelElement,ElementValueType]](sme.value, value):
        update_element_with_value(member, member_value)
    case model.File():
      assert isinstance(value, dict), f"FileValue must be a dict: {value}"
      sme.content_type = cast(model.ContentType, value.get('content_type'))
      assert sme.content_type is not None, "content_type is required"
      sme.value = cast(Optional[model.PathType], value.get('value'))
    case model.Range():
      assert isinstance(value, dict), f"RangeValue must be a dict: {value}"
      sme.min = cast(Optional[model.ValueDataType], value.get('min'))
      sme.max = cast(Optional[model.ValueDataType], value.get('max'))
    case model.MultiLanguageProperty():
      assert isinstance(value, dict), f"MultiLanguagePropertyValue must be a dict: {value}"
      sme.value = model.MultiLanguageTextType(cast(dict[str, str], value))
    case _:
      raise NotImplementedError(f"Unknown SubmodelElement type: {type(sme)}")


def from_json_object(value:Optional[ElementJsonValueType], proto:model.SubmodelElement) -> Optional[ElementValueType]:
  """Json 객체를 ElementValueType 객체로 변환한다.

  Args:
      value (Optional[JsonElementValueType]): 변환할 Json 객체.
      proto (model.SubmodelElement): 변환할 SubmodelElement 객체.

  Returns:
      Optional[ElementValueType]: 변환된 ElementValueType 객체.
  """
  if value is None:
    return None

  match proto:
    case model.Property():
      if isinstance(value, str):
        return cast(ElementValueType, model.datatypes.from_xsd(value, proto.value_type))
      else:
        return cast(model.ValueDataType, value)
    case model.SubmodelElementCollection():
      assert isinstance(value, dict)
      parsed_value = dict[str, Optional[ElementValueType]]()
      for member in proto.value:
        key = str(member.id_short)
        member_value = value.get(key)
        parsed_value[key] = from_json_object(member_value, member) if member_value is not None else None
      return parsed_value
    case model.SubmodelElementList():
      assert isinstance(value, list)
      return [ from_json_object(member_value, member) 
                for member, member_value in zip[tuple[model.SubmodelElement,ElementValueType]](proto.value, value) ]
    case model.File():
      assert isinstance(value, dict)
      ct, v = value.get('contentType'), value.get('value')
      assert ct is not None, "contentType is required"
      return cast(FileValue, { 'content_type': ct, 'value': v })
    case model.Range():
      assert isinstance(value, dict)
      min = cast(str, value.get('min'))
      min = model.datatypes.from_xsd(min, proto.value_type) if min is not None else None
      max = cast(str, value.get('max'))
      max = model.datatypes.from_xsd(max, proto.value_type) if max is not None else None
      return { 'min': min, 'max': max }
    case model.MultiLanguageProperty():
      assert isinstance(value, list)  # MultiLanguagePropertyJsonValue
      value_list = cast(list[dict[str, str]], value)
      return cast(MultiLanguagePropertyValue,
                  { lang:text for kv in value_list for lang, text in kv.items() })
    case _:
      raise NotImplementedError(f"Unknown SubmodelElement type: {type(proto)}")

def to_json_object(value:Optional[ElementValueType], proto:model.SubmodelElement) -> Optional[ElementJsonValueType]:
  if value is None:
    return None

  match proto:
    case model.Property():
      assert isinstance(value, model.ValueDataType)
      return model.datatypes.xsd_repr(value)
    case model.SubmodelElementCollection():
      assert isinstance(value, dict)
      if proto.value is None:
        return None
      collection_json_value = dict[str, Optional[ElementJsonValueType]]()
      for member in proto.value:
        key = str(member.id_short)
        member_value = value.get(key)
        if member_value is not None:
          collection_json_value[key] = to_json_object(member_value, member)
      return collection_json_value
    case model.SubmodelElementList():
      assert isinstance(value, list)
      return [ to_json_object(member_value, member) for member, member_value in zip[tuple[model.SubmodelElement,ElementValueType]](proto.value, value) ]
    case model.File():
      assert isinstance(value, dict)
      return cast(FileJsonValue, { 'contentType': value.get('content_type'), 'value': value.get('value') })
    case model.Range():
      assert isinstance(value, dict)
      min = cast(Optional[model.ValueDataType], value.get('min'))
      min = model.datatypes.xsd_repr(min) if min is not None else None
      max = cast(Optional[model.ValueDataType], value.get('max'))
      max = model.datatypes.xsd_repr(max) if max is not None else None
      return { 'min': min, 'max': max }
    case model.MultiLanguageProperty():
      assert isinstance(value, dict)
      return cast(MultiLanguagePropertyValue, { lang:text for lang, text in value.items() })
    case _:
      raise NotImplementedError(f"Unknown SubmodelElement type: {type(proto)}")


from __future__ import annotations

from typing import Self, Any, cast, TypeVar, Generic

import json
from datetime import datetime, timedelta

from basyx.aas import model

T = TypeVar('T')


class JsonSerializable(Generic[T]):
  @classmethod
  def from_json_object(cls, data: Any) -> T: ...

  def to_json_object(self) -> Any: ...

  @classmethod
  def from_json(cls, json_str: str) -> T:
    json_obj = json.loads(json_str)
    return cls.from_json_object(json_obj)

  def to_json(self) -> str:
    return json.dumps(self.to_json_object())


def datetime_to_iso8601(dt: datetime) -> str:
    """Convert a datetime object to ISO8601 datetime string format."""
    return dt.isoformat()

def iso8601_to_datetime(iso8601: str) -> datetime:
    # 밀리초 부분이 3자리가 아닌 경우 처리
    if '.' in iso8601:
        base, ms = iso8601.split('.')
        # 밀리초 부분을 3자리로 맞춤
        ms = ms.ljust(3, '0')
        iso8601 = f"{base}.{ms}"
    return datetime.fromisoformat(iso8601)

def timedelta_to_iso8601(delta: timedelta) -> str:
    return second_to_iso8601(delta.total_seconds())

def second_to_iso8601(total_seconds: float) -> str:
    """Convert a timedelta object to ISO8601 duration string format."""
    days, remainder = divmod(total_seconds, 86400)
    hours, remainder = divmod(remainder, 3600)
    minutes, remainder = divmod(remainder, 60)
    seconds, milliseconds = divmod(remainder, 1)
    
    parts = []
    if days > 0:
        parts.append(f"{int(days)}D")
    if hours > 0 or minutes > 0 or seconds > 0 or milliseconds > 0 or (not parts):
        time_part = "T"
        if hours > 0:
            time_part += f"{int(hours)}H"
        if minutes > 0:
            time_part += f"{int(minutes)}M"
        if seconds > 0 or milliseconds > 0 or (not parts and time_part == "T"):
            seconds_str = f"{int(seconds)}"
            if milliseconds > 0:
                seconds_str += f".{int(milliseconds * 1000):03d}"
            time_part += f"{seconds_str}S"
        parts.append(time_part)
    
    return "P" + "".join(parts)

def iso8601_to_timedelta(iso8601: str) -> timedelta:
    import isodate
    return isodate.parse_duration(iso8601)

def to_nonnull(value: T|None) -> T:
  assert value is not None
  return value

from dateutil.relativedelta import relativedelta

def to_str(sme:model.SubmodelElement) -> str|None:
  value = cast(model.Property, sme).value
  return str(value) if value is not None else None
def to_int(sme:model.SubmodelElement) -> int|None:
  value = cast(model.Property, sme).value
  return model.datatypes.Int(value) if value is not None else None
def to_datetime(sme:model.SubmodelElement) -> datetime|None:
  return sme.value if sme.value is not None else None   # type: ignore
def to_duration(sme:model.SubmodelElement) -> relativedelta|None:
  return sme.value if sme.value is not None else None   # type: ignore
def to_mlstr(mlp: model.SubmodelElement) -> str|None:
  try:
    if mlt := mlp.value:    # type: ignore
      tup = next(iter(mlt.items()))
      return tup[1] if tup is not None else None
    else:
      return None
  except StopIteration:
    return None



from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta

def relativedelta_to_timedelata(rd:relativedelta, base:datetime) -> timedelta:
  return (base + rd) - base if rd else timedelta(seconds=0)
def relativedelta_to_seconds(rd:relativedelta, base:datetime) -> float:
  return relativedelta_to_timedelata(rd, base).total_seconds()

def timedelta_to_relativedelta(td:timedelta) -> relativedelta:
  return relativedelta(days=td.days,
                        seconds=td.seconds,
                        microseconds=td.microseconds)


from __future__ import annotations

    
class MDTException(Exception):
  def __init__(self, details:str) -> None:
    self.details = details
    super().__init__(details)
    
  def __str__(self) -> str:
    return repr(self)


class InternalError(MDTException):
  def __init__(self, details:str) -> None:
    super().__init__(details)


class TimeoutError(MDTException):
  def __init__(self, details:str) -> None:
    super().__init__(details)


class CancellationError(MDTException):
  def __init__(self, details:str) -> None:
    super().__init__(details)


class OperationError(MDTException):
  def __init__(self, details:str) -> None:
    super().__init__(details)


class RemoteError(MDTException):
  def __init__(self, details:str) -> None:
    super().__init__(details)

import requests
class MDTInstanceConnectionError(MDTException):
  def __init__(self, details:str, cause:requests.exceptions.ConnectionError) -> None:
    super().__init__(details)
    self.cause = cause

  def __repr__(self) -> str:
    return f"{self.__class__.__name__}(details={self.details}, cause={self.cause})"

   
class ResourceAlreadyExistsError(MDTException):
  def __init__(self, details:str) -> None:
    super().__init__(details)
      
  @classmethod
  def create(cls, resource_type:str, id_spec:str):
    return ResourceAlreadyExistsError(f"Resource(type={resource_type}, {id_spec})")
    
    
class ResourceNotFoundError(MDTException):
  def __init__(self, details:str) -> None:
    super().__init__(details)
      
  @classmethod
  def create(cls, resource_type:str, id_spec:str):
    return ResourceNotFoundError(f"Resource(type={resource_type}, {id_spec})")
        
    
class InvalidResourceStateError(MDTException):
  def __init__(self, details:str) -> None:
    super().__init__(details)
      
  @classmethod
  def create(cls, resource_type:str, id_spec:str, status):
    return InvalidResourceStateError(f"Resource(type={resource_type}, {id_spec}), status={status}")


from __future__ import annotations

from typing import Any, Iterator, Generator, Optional
from abc import ABC, abstractmethod

import requests
from urllib.parse import quote

from .submodel import SubmodelService, SubmodelServiceCollection
from .reference import DefaultElementReference
from .descriptor import InstanceDescriptor, MDTParameterDescriptor, MDTSubmodelDescriptor, MDTOperationDescriptor, MDTAssetType, AssetKind, MDTInstanceStatus
from .parameter import MDTParameterCollection, MDTParameter
from .timeseries import TimeSeriesService
from .operation import OperationSubmodelService
from .http_client import parse_response, parse_list_response, parse_none_response
from .exceptions import InvalidResourceStateError


mdt_endpoint:Optional[str] = None
mdt_manager:Optional[MDTInstanceManager] = None

def connect(endpoint: str) -> MDTInstanceManager:
  global mdt_endpoint, mdt_manager
  mdt_endpoint = endpoint
  mdt_manager = MDTInstanceManager(endpoint)
  return mdt_manager


class MDTInstanceManager:
  def __init__(self, endpoint: str):
    self.__endpoint = endpoint
    self.__instances = MDTInstanceCollection(endpoint)

  @property
  def endpoint(self) -> str:
    return self.__endpoint

  @property
  def instances(self) -> MDTInstanceCollection:
    return self.__instances

  def resolve_reference(self, ref_string: str) -> DefaultElementReference:
    parts = ref_string.split(':')
    match parts[0]:
      case 'param':
        assert len(parts) == 3, f"Invalid parameter reference: {ref_string}"
        return self.instances[parts[1]].parameters[parts[2]]
      case 'oparg':
        assert len(parts) == 5, f"Invalid operation argument reference: {ref_string}"
        op = self.instances[parts[1]].operations[parts[2]]
        match parts[3]:
          case 'in':
            return op.input_arguments[parts[4]]
          case 'out':
            return op.output_arguments[parts[4]]
          case _:
            raise ValueError(f"Invalid operation argument reference: {ref_string}")
      case _:
        url = f"{self.__endpoint}/references/$url?ref={quote(ref_string)}"
        resp = requests.get(url)
        return DefaultElementReference(ref_string=ref_string, endpoint=parse_response(resp))


class MDTInstanceCollection:
  def __init__(self, endpoint: str):
    self.__endpoint = endpoint
    self.__url_prefix = f"{endpoint}/instances"

  def __bool__(self) -> bool:
    return len(self) > 0

  def __len__(self) -> int:
    resp = requests.get(self.__url_prefix)
    return len(parse_list_response(resp, InstanceDescriptor))
        
  def __iter__(self) -> Iterator[MDTInstance]:
      resp = requests.get(self.__url_prefix)
      inst_desc_list = parse_list_response(resp, InstanceDescriptor)
      return iter(MDTInstance(inst_desc, self.__endpoint) for inst_desc in inst_desc_list)
  
  def __contains__(self, key:str) -> bool:
      url = f'{self.__url_prefix}/{key}'
      resp = requests.get(url)
      return resp.status_code == 200
        
  def __getitem__(self, key:str) -> MDTInstance:
      url = f'{self.__url_prefix}/{key}'
      resp = requests.get(url)
      inst_desc: InstanceDescriptor = parse_response(resp, InstanceDescriptor)  # type: ignore
      return MDTInstance(inst_desc, self.__endpoint)
    
  def find(self, condition:str) -> Generator[MDTInstance, None, None]:
      resp = requests.get(self.__url_prefix, params={'filter': f"{condition}"})
      inst_desc_list = parse_list_response(resp, InstanceDescriptor)
      return (MDTInstance(inst_desc, self.__endpoint) for inst_desc in inst_desc_list)
    
  def add(self, id:str, port:int, inst_dir:str) -> MDTInstance:
    import shutil
    shutil.make_archive(inst_dir, 'zip', inst_dir)
    zipped_file = f'{inst_dir}.zip'
    
    from requests_toolbelt.multipart.encoder import MultipartEncoder
    m = MultipartEncoder(
      fields = {
        'id': id,
        'port': str(port),
        'bundle': ('filename', open(zipped_file, 'rb'), 'application/zip')
      }
    )
    resp = requests.post(self.__url_prefix, data=m, headers={'Content-Type': m.content_type}, verify=False)
    inst_desc: InstanceDescriptor = parse_response(resp, InstanceDescriptor)  # type: ignore 
    return MDTInstance(inst_desc, self.__endpoint)
    
  def __delitem__(self, key:str) -> None:
    url = f'{self.__url_prefix}/{key}'
    resp = requests.delete(url)
    parse_none_response(resp)

  def remove(self, id:str) -> None:
    url = f'{self.__url_prefix}/{id}'
    resp = requests.delete(url)
    return parse_none_response(resp)
        
  def remove_all(self) -> None:
    url = f"{self.__url_prefix}"
    resp = requests.delete(url)
    parse_none_response(resp)


class MDTInstance:
  def __init__(self, descriptor:InstanceDescriptor, endpoint: str) -> None:
    self.__descriptor = descriptor
    self.__instanceUrl = f"{endpoint}/instances/{descriptor.id}"
    self.__submodel_descriptors = None

  @property
  def id(self) -> str:
    return self.__descriptor.id

  @property
  def aas_id(self) -> str:
    return self.__descriptor.aas_id

  @property
  def aas_id_short(self) -> str|None:
    return self.__descriptor.aas_id_short

  @property
  def global_asset_id(self) -> str|None:
    return self.__descriptor.global_asset_id

  @property
  def asset_type(self) -> MDTAssetType|None:
    return self.__descriptor.asset_type

  @property
  def asset_kind(self) -> AssetKind|None:
    return self.__descriptor.asset_kind

  @property
  def status(self) -> MDTInstanceStatus:
    return self.__descriptor.status

  @property
  def base_endpoint(self) -> str|None:
    return self.__descriptor.base_endpoint

  def is_running(self) -> bool:
    return self.status == MDTInstanceStatus.RUNNING

  @property
  def parameters(self) -> MDTParameterCollection:
    if not self.is_running():
      raise InvalidResourceStateError.create("MDTInstance", f"id={self.id}", self.status)

    url = f"{self.__instanceUrl}/model/parameters"
    resp = requests.get(url)
    desc_list = parse_list_response(resp, MDTParameterDescriptor)
    return MDTParameterCollection([MDTParameter(desc) for desc in desc_list])

  @property
  def submodel_descriptors(self) -> dict[str, MDTSubmodelDescriptor]:
    if not self.is_running():
      raise InvalidResourceStateError.create("MDTInstance", f"id={self.id}", self.status)

    url = f"{self.__instanceUrl}/model/submodels"
    resp = requests.get(url)
    return { desc.id_short:desc for desc in parse_list_response(resp, MDTSubmodelDescriptor) }

  @property
  def operation_descriptors(self) -> dict[str, MDTOperationDescriptor]:
    if not self.is_running():
      raise InvalidResourceStateError.create("MDTInstance", f"id={self.id}", self.status)

    url = f"{self.__instanceUrl}/model/operations"
    resp = requests.get(url)
    return { desc.id:desc for desc in parse_list_response(resp, MDTOperationDescriptor) }

  @property
  def submodel_services(self) -> SubmodelServiceCollection:
    return SubmodelServiceCollection[SubmodelService](self, self.submodel_descriptors)

  @property
  def operations(self) -> SubmodelServiceCollection[OperationSubmodelService]:
    op_sm_desc_dict = { id:sm_desc for id, sm_desc in self.submodel_descriptors.items() \
                                        if sm_desc.is_simulation() or sm_desc.is_ai() }
    return SubmodelServiceCollection[OperationSubmodelService](self, op_sm_desc_dict)

  @property
  def timeseries(self) -> SubmodelServiceCollection[TimeSeriesService]:
    ts_sm_desc_dict = { id:sm_desc for id, sm_desc in self.submodel_descriptors.items() if sm_desc.is_time_series() }
    return SubmodelServiceCollection[TimeSeriesService](self, ts_sm_desc_dict)

  def start(self, nowait=False) -> None:
    url = f"{self.__instanceUrl}/start"
    resp = requests.put(url, data="")

    self.reload_descriptor()
    if nowait:
      if self.__descriptor.status != MDTInstanceStatus.STARTING and MDTInstanceStatus.RUNNING:
        raise InvalidResourceStateError.create("MDTInstance", f"id={self.id}", self.__descriptor.status)
    else:
      poller = InstanceStartPoller(f"{self.__instanceUrl}", init_desc=self.__descriptor)
      poller.wait_for_done()
      self.__descriptor = poller.desc
      if self.__descriptor.status != MDTInstanceStatus.RUNNING:
        raise InvalidResourceStateError.create("MDTInstance", f"id={self.id}", self.__descriptor.status)
  
  def stop(self, nowait=False) -> None:
    url = f"{self.__instanceUrl}/stop"
    resp = requests.put(url, data="")
    parse_none_response(resp)

    self.reload_descriptor()
    if nowait:
      if self.__descriptor.status != MDTInstanceStatus.STOPPING and MDTInstanceStatus.STOPPED:
        raise InvalidResourceStateError.create("MDTInstance", f"id={self.id}", self.__descriptor.status)
    else:
      poller = InstanceStopPoller(f"{self.__instanceUrl}", init_desc=self.__descriptor)
      poller.wait_for_done()
      self.__descriptor = poller.desc
      if self.__descriptor.status != MDTInstanceStatus.STOPPED:
        raise InvalidResourceStateError.create("MDTInstance", f"id={self.id}", self.__descriptor.status)
    
  def reload_descriptor(self) -> InstanceDescriptor:
    resp = requests.get(self.__instanceUrl)
    self.__descriptor: InstanceDescriptor = parse_response(resp, InstanceDescriptor)  # type: ignore
    return self.__descriptor

  def __str__(self) -> str:
    return f"HttpMDTInstance({self.__descriptor})"

  def __repr__(self) -> str:
    return self.__str__()


import time
class StatusPoller(ABC):
  """
  Abstract base class for polling the status of an operation.
  Attributes:
    poll_interval (float): The interval in seconds between each poll.
    timeout (Optional[float]): The maximum time in seconds to wait for the operation to complete. If None, wait indefinitely.
  Methods:
    is_done() -> bool:
        Abstract method to check if the operation is done. Must be implemented by subclasses.
    wait_for_done() -> None:
      Waits for the operation to complete by repeatedly calling `check_done` at intervals specified by `poll_interval`.
      Raises:
        TimeoutError: If the operation does not complete within the specified timeout.
  """
  def __init__(self, poll_interval:float, timeout:float|None=None):
      self.poll_interval = poll_interval
      self.timeout = timeout
      
  @abstractmethod
  def is_done(self) -> bool: pass
  
  def wait_for_done(self) -> None:
      # 타임아웃 (self.timeout)이 있는 경우 최종 제한 시간을 계산하고,    
      # 타임아웃이 없는 경우 due를 None으로 설정하여 무제한 대기하도록 한다.
      started = time.time()
      due = started + self.timeout if self.timeout else None
      # 다음 폴링 시간을 계산한다.
      next_wakeup = started + self.poll_interval
      
      while not self.is_done():
          now = time.time()
          
          # 타임 아웃까지 남은 시간이 일정 시간 이내인 경우에는 TimeoutError를 발생시킨다.
          # 그렇지 않은 경우는 다음 폴링 시간까지 대기한다.
          if due and (due - now) < 0.01:
              raise TimeoutError(f'timeout={self.timeout}')
          
          # 다음 폴링 시간까지 남은 시간이 짧으면 대기하지 않고 바로 다음 폴링 시도한다.
          sleep_time = next_wakeup - now
          if sleep_time > 0.001:
              time.sleep(sleep_time)
          next_wakeup += self.poll_interval

class InstanceStartPoller(StatusPoller):
    def __init__(self, status_url:str, init_desc:InstanceDescriptor|None=None,
                 poll_interval:float=1.0, timeout:float|None=None) -> None:
        super().__init__(poll_interval=poll_interval, timeout=timeout)
        self.status_url = status_url
        self.desc = init_desc
        
    def is_done(self) -> bool:
        if self.desc.status == MDTInstanceStatus.STARTING:
            resp = requests.get(self.status_url)
            self.desc: InstanceDescriptor = parse_response(resp, InstanceDescriptor)  # type: ignore
            return self.desc.status != MDTInstanceStatus.STARTING
        else:
            return True
    
class InstanceStopPoller(StatusPoller):
  def __init__(self, status_url:str, init_desc:InstanceDescriptor|None=None,
                poll_interval:float=1.0, timeout:float|None=None) -> None:
    super().__init__(poll_interval=poll_interval, timeout=timeout)
    self.status_url = status_url
    self.desc = init_desc
        
  def is_done(self) -> bool:
    resp = requests.get(self.status_url)
    self.desc: InstanceDescriptor = parse_response(resp, InstanceDescriptor)  # type: ignore
    return self.desc.status == MDTInstanceStatus.STOPPED


from __future__ import annotations

from typing import Any, cast, Optional, Iterable
from enum import Enum, auto

import json
import datetime
from dataclasses import dataclass, field
from dataclass_wizard import JSONWizard

from basyx.aas import model
from .basyx import serde as basyx_serde
from . import utils

    
class SecurityTypeEnum(Enum):
  NONE = auto()
  RFC_TLSA = auto()
  W3C_DID = auto()

@dataclass(slots=True)
class SecurityAttributeObject(JSONWizard):
  type: SecurityTypeEnum
  key: str
  value: str

@dataclass(slots=True)
class ProtocolInformation:
  href: Optional[str]
  endpointProtocol: Optional[str] = field(default=None)
  endpointProtocolVersion: Optional[str] = field(default=None)
  subprotocol: Optional[str] = field(default=None)
  subprotocolBody: Optional[str] = field(default=None)
  subprotocolBody_encoding: Optional[str] = field(default=None)
  securityAttributes: list[SecurityAttributeObject] = field(default_factory=list)

@dataclass(slots=True)
class Endpoint:
  interface: str
  protocolInformation: ProtocolInformation

@dataclass(slots=True)
class OperationVariable:
  value: model.SubmodelElement

  @classmethod
  def from_dict(cls, data: dict) -> OperationVariable:
    return cls(value=basyx_serde.from_dict(data['value']))

  def to_dict(self) -> dict[str, Any]:
    return { 'value': json.loads(basyx_serde.to_json(self.value)) }

@dataclass(slots=True)
class OperationResult:
  messages: Optional[list[str]]
  execution_state: str
  success: bool
  output_arguments: Optional[list[OperationVariable]]
  inoutput_arguments: Optional[list[OperationVariable]]

  @classmethod
  def from_dict(cls, data: dict) -> OperationResult:
    output_arguments = data.get('outputArguments')
    if output_arguments:
      output_arguments = [OperationVariable.from_dict(arg) for arg in output_arguments]
    inoutput_arguments = data.get('inoutputArguments')
    if inoutput_arguments:
      inoutput_arguments = [OperationVariable.from_dict(arg) for arg in inoutput_arguments]

    return cls(messages = data.get('messages'),
              execution_state = data['executionState'],
              success = data['success'],
              output_arguments=output_arguments, 
              inoutput_arguments=inoutput_arguments)
  
  @classmethod
  def from_json(cls, json_str: str) -> OperationResult:
    return cls.from_dict(json.loads(json_str))

@dataclass(slots=True)
class OperationHandle:
  handle_id: str

  @classmethod
  def from_json(cls, json_str: str) -> OperationHandle:
    json_dict = json.loads(json_str)
    return cls(handle_id=json_dict['handleId'])
    
@dataclass(slots=True)
class OperationRequest:
  input_arguments: Iterable[OperationVariable]
  inoutput_arguments: Iterable[OperationVariable]
  client_timeout_duration: datetime.timedelta

  def to_json(self) -> str:
    in_opv_list = [ op_var.to_dict() for op_var in self.input_arguments ]
    inout_opv_list = [ op_var.to_dict() for op_var in self.inoutput_arguments ]
    return json.dumps({
      'inputArguments': in_opv_list,
      # 'inoutputArguments': inout_opv_list,
      'clientTimeoutDuration': utils.timedelta_to_iso8601(self.client_timeout_duration)
    })



__version__ = "0.1.16"

# SSL 인증서 검증 경고 억제
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

from .instance import *
from .value import *
from .descriptor import *
from .exceptions import *
from .timeseries import *
from .reference import *
from .operation import *
from . import aas_misc as aas
from . import basyx
from . import airflow



from __future__ import annotations

from typing import Optional
from dataclasses import dataclass, field
from enum import Enum

from dataclass_wizard import JSONWizard


class MDTInstanceStatus(Enum):
  STOPPED = "STOPPED"
  STARTING = "STARTING"
  RUNNING = "RUNNING"
  STOPPING = "STOPPING"
  FAILED = "FAILED"

class MDTAssetType(Enum):
  Machine = "Machine"
  Process = "Process"
  Line = "Line"
  Factory = "Factory"

class AssetKind(Enum):
  INSTANCE = "INSTANCE"
  NOT_APPLICABLE = "NOT_APPLICABLE"
  TYPE = "TYPE"


@dataclass(frozen=True, slots=True)
class InstanceDescriptor(JSONWizard):
  """
  MDTInstance Descriptor

  Attributes:
  ----------
    id: str
      The unique identifier of the MDTInstance.
    status: MDTInstanceStatus
      The Status of the MDTInstance.
    base_endpoint: Optional[str]
      The Base Endpoint of the MDTInstance.
    aas_id: str
      The AAS ID of the MDTInstance.
    aas_idshort: Optional[str]
      The AAS ID Short of the MDTInstance.
    global_asset_id: Optional[str]
      The Global Asset ID of the MDTInstance.
    asset_type: Optional[MDTAssetType]
      The Asset Type of the MDTInstance.
    asset_kind: Optional[AssetKind]
      The Asset Kind of the MDTInstance.
  """
  id: str
  status: MDTInstanceStatus
  aas_id: str
  base_endpoint: Optional[str] = field(default=None, hash=False, compare=False)
  aas_id_short: Optional[str] = field(default=None, hash=False, compare=False)
  global_asset_id: Optional[str] = field(default=None, hash=False, compare=False)
  asset_type: MDTAssetType|None = field(default=None, hash=False, compare=False)
  asset_kind: AssetKind|None = field(default=None, hash=False, compare=False)

@dataclass(frozen=True, slots=True)
class MDTParameterDescriptor(JSONWizard):
  id: str
  value_type: str
  reference: str
  name: Optional[str] = None
  endpoint: Optional[str] = None


SEMANTIC_ID_INFOR_MODEL_SUBMODEL = "https://etri.re.kr/mdt/Submodel/InformationModel/1/1"
SEMANTIC_ID_AI_SUBMODEL = "https://etri.re.kr/mdt/Submodel/AI/1/1"
SEMANTIC_ID_SIMULATION_SUBMODEL = "https://etri.re.kr/mdt/Submodel/Simulation/1/1"
SEMANTIC_ID_DATA_SUBMODEL = "https://etri.re.kr/mdt/Submodel/Data/1/1"
SEMANTIC_ID_TIME_SERIES_SUBMODEL = 'https://admin-shell.io/idta/TimeSeries/1/1'

@dataclass(frozen=True, slots=True)
class MDTSubmodelDescriptor(JSONWizard):
  id: str
  id_short: str
  semantic_id: str
  endpoint: Optional[str]

  def is_information_model(self) -> bool:
    return self.semantic_id == SEMANTIC_ID_INFOR_MODEL_SUBMODEL

  def is_data(self) -> bool:
    return self.semantic_id == SEMANTIC_ID_DATA_SUBMODEL

  def is_simulation(self) -> bool:
    return self.semantic_id == SEMANTIC_ID_SIMULATION_SUBMODEL

  def is_ai(self) -> bool:
    return self.semantic_id == SEMANTIC_ID_AI_SUBMODEL

  def is_time_series(self) -> bool:
    return self.semantic_id == SEMANTIC_ID_TIME_SERIES_SUBMODEL


@dataclass(frozen=True, slots=True)
class ArgumentDescriptor(JSONWizard):
  id: str
  id_short_path: str
  value_type: str
  reference: str


@dataclass(frozen=True, slots=True)
class MDTOperationDescriptor(JSONWizard):
  id: str
  operation_type: str
  input_arguments: list[ArgumentDescriptor]
  output_arguments: list[ArgumentDescriptor]



from __future__ import annotations

from typing import Optional, cast, TypeVar
from abc import ABC, abstractmethod
from collections import UserDict
from functools import cached_property

import requests
import json
from urllib.parse import quote

from basyx.aas import model

from . import fa3st
from .basyx import serde as basyx_serde
from .value import ElementValueType, from_json_object, to_json_object
from .http_client import parse_none_response
from .exceptions import ResourceNotFoundError


def reference(ref_string: str) -> LazyElementReference:
  return LazyElementReference(ref_string)

class ElementReference(ABC):
  @abstractmethod
  def read_json(self) -> str:
    """ElementReference가 가리키는 SubmodelElement 객체를 json 문자열로 반환한다.

    Returns:
        str: ElementReference가 가리키는 SubmodelElement 객체를 json 문자열로 반환한다.
    """
    pass

  @abstractmethod
  def write_json(self, json_str: str) -> None:
    """SubmodelElement JSON 문자열을 SubmodelElement 객체로 변환하여
    ElementReference가 가리키는 SubmodelElement 객체를 변경한다.

    Parameters:
        json_str (str): 저장할 json 문자열.
    """
    pass

  def read(self) -> model.SubmodelElement:
    """ElementReference가 가리키는 SubmodelElement 객체를 반환한다.

    Returns:
        model.SubmodelElement: ElementReference가 가리키는 SubmodelElement 객체.
    """
    return basyx_serde.from_json(self.read_json())

  def write(self, sme: model.SubmodelElement) -> None:
    """주어진 SubmodelElement를 이용해서 ElementReference가 가리키는 SubmodelElement 객체를 변경한다.

    Parameters:
        sme (model.SubmodelElement): 변경할 SubmodelElement 객체.
    """
    self.write_json(basyx_serde.to_json(sme))
  
  @abstractmethod
  def read_value(self) -> Optional[ElementValueType]:
    """ElementReference가 가리키는 SubmodelElement 객체의 값을 반환한다.

    Returns:
        Optional[ElementValueType]: ElementReference가 가리키는 SubmodelElement 객체의 값.
    """
    pass

  @abstractmethod
  def update_value(self, value: Optional[ElementValueType]) -> None:
    """ElementReference가 가리키는 SubmodelElement 객체의 값을 변경한다.

    Parameters:
        value (Optional[ElementValueType]): 변경할 값.
    """
    pass

  @abstractmethod
  def update_value_with_json_string(self, json_str: str) -> None:
    """ElementReference가 가리키는 SubmodelElement 객체의 값을 json 문자열로 변경한다.

    Parameters:
        json_str (str): 변경할 json 문자열.
    """
    pass


class DefaultElementReference(ElementReference):
  def __init__(self, ref_string: str, endpoint: str) -> None:
    assert ref_string is not None, f"ref_string is None for ElementReference"
    assert endpoint is not None, f"element_url is None for ElementReference[{ref_string}]"

    self.__ref_string = ref_string
    self.__endpoint = endpoint

  @property
  def ref_string(self) -> str:
    return self.__ref_string

  @property
  def endpoint(self) -> str:
    return self.__endpoint

  @cached_property
  def prototype(self) -> model.SubmodelElement:
    return cast(model.SubmodelElement, fa3st.call_get(self.endpoint,
                                                      deserializer=basyx_serde.from_json))

  @property
  def id_short(self) -> str:
    """ElementReference가 가리키는 SubmodelElement 객체의 idShort 값을 반환한다.

    Returns:
        str: ElementReference가 가리키는 SubmodelElement 객체의 idShort 값.
    """
    return str(self.prototype.id_short)

  @property
  def model_type(self) -> type[model.SubmodelElement]:
    """ElementReference가 가리키는 SubmodelElement 객체의 모델 타입을 반환한다.

    Returns:
        str: ElementReference가 가리키는 SubmodelElement 객체의 모델 타입.
    """
    return type(self.prototype)

  @property
  def value_type(self) -> Optional[str]:
    """ElementReference가 가리키는 SubmodelElement 객체의 값 타입을 반환한다.
    Property 타입의 SubmodelElement 객체의 경우에만 사용할 수 있고, 그 외의 경우에는 None을 반환한다.

    Returns:
        Optional[str]: ElementReference가 가리키는 SubmodelElement 객체의 값 타입.
    """
    if isinstance(self.prototype, model.Property):
      return model.datatypes.XSD_TYPE_NAMES[self.prototype.value_type]
    else:
      return None

  @property
  def semantic_id(self) -> Optional[model.Reference]:
    """ElementReference가 가리키는 SubmodelElement 객체의 semanticId 값을 반환한다.

    Returns:
        Optional[model.Reference]: ElementReference가 가리키는 SubmodelElement 객체의 semanticId 값.
    """
    return self.prototype.semantic_id

  def pathes(self) -> list[str]:
    """ElementReference가 가리키는 SubmodelElement 객체의 경로를 반환한다.

    Returns:
        list[str]: ElementReference가 가리키는 SubmodelElement 객체의 경로.
    """
    path_list = cast(str, fa3st.call_get(f"{self.endpoint}/$path"))
    path_list = [ path.strip() for path in path_list[2:-2].split(',') ]
    return [ path[1:-1] for path in path_list ]

  def exists(self) -> bool:
    """ElementReference가 가리키는 SubmodelElement 객체가 존재하는지 여부를 반환한다.

    Returns:
        bool: ElementReference가 가리키는 SubmodelElement 객체가 존재하는지 여부.
    """
    try:
      self.prototype
      return True
    except ResourceNotFoundError:
      return False

  def read_json(self) -> str:
    """ElementReference가 가리키는 SubmodelElement 객체를 json 문자열로 반환한다.

    Returns:
        str: ElementReference가 가리키는 SubmodelElement 객체를 json 문자열로 반환한다.
    """
    return str(fa3st.call_get(self.endpoint))

  def write_json(self, json_str: str) -> None:
    """SubmodelElement JSON 문자열을 SubmodelElement 객체로 변환하여
    ElementReference가 가리키는 SubmodelElement 객체를 변경한다.

    Parameters:
        json_str (str): 저장할 json 문자열.
    """
    fa3st.call_put(self.endpoint, json_str)

  def read(self) -> model.SubmodelElement:
    """ElementReference가 가리키는 SubmodelElement 객체를 반환한다.

    Returns:
        model.SubmodelElement: ElementReference가 가리키는 SubmodelElement 객체.
    """
    return cast(model.SubmodelElement, fa3st.call_get(self.endpoint,
                                                      deserializer=basyx_serde.from_json))

  def write(self, sme: model.SubmodelElement) -> None:
    """ElementReference가 가리키는 SubmodelElement 객체를 변경한다.

    Parameters:
        sme (model.SubmodelElement): 변경할 SubmodelElement 객체.
    """
    fa3st.call_put(self.endpoint, basyx_serde.to_json(sme))
  
  def read_value(self) -> Optional[ElementValueType]:
    """ElementReference가 가리키는 SubmodelElement 객체의 값을 반환한다.

    Returns:
        Optional[ElementValueType]: ElementReference가 가리키는 SubmodelElement 객체의 값.
    """
    json_str = cast(str, fa3st.call_get(f"{self.endpoint}/$value"))
    json_obj = json.loads(json_str)
    _, value = next(iter(json_obj.items()))
    return from_json_object(value, self.prototype)

  def update_value(self, value: Optional[ElementValueType]) -> None:
    """ElementReference가 가리키는 SubmodelElement 객체의 값을 변경한다.

    Parameters:
        value (Optional[ElementValueType]): 변경할 값.
    """
    value_json_str = json.dumps(to_json_object(value, self.prototype))
    fa3st.call_patch(f"{self.endpoint}/$value", value_json_str)

  def update_value_with_json_string(self, json_str: str) -> None:
    """ElementReference가 가리키는 SubmodelElement 객체의 값을 json 문자열로 변경한다.

    Parameters:
        json_str (str): 변경할 json 문자열.
    """
    resp = requests.put(f"{self.endpoint}/$value", json=json_str)
    parse_none_response(resp)

  def add(self, sme: model.SubmodelElement) -> None:
    """ElementReference가 가리키는 SubmodelElement 객체를 추가한다.

    Parameters:
        sme (model.SubmodelElement): 추가할 SubmodelElement 객체.
    """
    fa3st.call_post(self.endpoint, basyx_serde.to_json(sme))

  def remove(self) -> None:
    """ElementReference가 가리키는 SubmodelElement 객체를 삭제한다.

    Parameters:
        sme (model.SubmodelElement): 삭제할 SubmodelElement 객체.
    """
    fa3st.call_delete(self.endpoint)

  def get_attachment(self) -> Optional[bytes]:
    resp = requests.get(url=f'{self.endpoint}/attachment', verify=False)
    ret = fa3st.read_file_response(resp)
    return ret[1] if ret else None

  def put_attachment(self, file_path:str, content_type:Optional[str]=None) -> None:
    import os
    from .value import to_file_value
    from requests_toolbelt.multipart.encoder import MultipartEncoder

    file_value = to_file_value(file_path, content_type=content_type)
    self.update_value(file_value)
    
    url = f'{self.endpoint}/attachment'
    file_name = os.path.basename(file_path)
    content_type = file_value['content_type']
    m = MultipartEncoder(
      fields={
        'fileName': file_name,
        'contentType': content_type,
        'content': ('filename', open(file_path, 'rb'), content_type)
      }
    )
    requests.put(url, data=m, headers={'Content-Type': m.content_type}, verify=False)
      
  def delete_attachment(self) -> None:
    file_sme = (cast(model.File, self.prototype))
    requests.delete(f'{self.endpoint}/attachment', verify=False)
    file_sme.value = None
    self.write(file_sme)

  def __repr__(self) -> str:
    return self.ref_string


T = TypeVar('T', bound=ElementReference)
class ElementReferenceDict(UserDict[str,T], ElementReference):
  def __init__(self, ref_dict: dict[str,T]):
    super().__init__(ref_dict)

  def read(self) -> model.SubmodelElementCollection:
    raise NotImplementedError("ElementReferenceDict.read() is not implemented")

  def write(self, sme: model.SubmodelElementCollection) -> None:
    raise NotImplementedError("ElementReferenceDict.write() is not implemented")

  def read_json(self) -> str:
    raise NotImplementedError("ElementReferenceDict.read_json() is not implemented")

  def write_json(self, json_str: str) -> None:
    raise NotImplementedError("ElementReferenceDict.write_json() is not implemented")

  def read_value(self) -> dict[str, Optional[ElementValueType]]:
    return { k:ref.read_value() for k, ref in self.data.items() }

  def update_value(self, new_values:dict[str, Optional[ElementValueType]]) -> None:
    for k,v in new_values.items():
      if k in self.data:
        self.data[k].update_value(v)
      else:
        raise ResourceNotFoundError.create("ElementReference", f"key={k}")

  def update_value_with_json_string(self, json_str: str) -> None:
    json_obj = json.loads(json_str)
    assert isinstance(json_obj, dict), f"Invalid json object: {json_str}"

    for k, v in json_obj.items():
      if k in self.data:
        self.data[k].update_value_with_json_string(json.dumps(v))


class LazyElementReference(ElementReference):
  def __init__(self, ref_string: str) -> None:
    self.ref_string = ref_string

  @cached_property
  def reference(self) -> DefaultElementReference:
    from .instance import mdt_manager
    assert mdt_manager is not None, "MDTManager is not connected"
    return mdt_manager.resolve_reference(self.ref_string)

  def read_json(self) -> str:
    return self.reference.read_json()

  def write_json(self, json_str: str) -> None:
    self.reference.write_json(json_str)

  def read(self) -> model.SubmodelElement:
    return self.reference.read()

  def write(self, sme: model.SubmodelElement) -> None:
    self.reference.write(sme)

  def read_value(self) -> Optional[ElementValueType]:
    return self.reference.read_value()

  def update_value(self, value: Optional[ElementValueType]) -> None:
    self.reference.update_value(value)

  def update_value_with_json_string(self, json_str: str) -> None:
    self.reference.update_value_with_json_string(json_str)


from __future__ import annotations

import requests

from .exceptions import MDTException, RemoteError, ResourceNotFoundError


def parse_none_response(resp:requests.Response) -> None:
    if resp.status_code >= 200 and resp.status_code < 300:
        return
    else:
        raise to_exception(resp)

def parse_response(resp:requests.Response, result_cls:type|None=None):
  if resp.status_code >= 200 and resp.status_code < 300:
    content_type = resp.headers['content-type']
    if content_type == 'application/json':
      json = resp.json()
      return result_cls.from_dict(json) if result_cls else json
    elif content_type.startswith('text/plain'):
      return resp.text
    else:
      raise MDTException(f"Unsupported content type: {content_type}")
  else:
    raise to_exception(resp)

def parse_list_response(resp:requests.Response, result_cls:type=None):
  if resp.status_code >= 200 and resp.status_code < 300:
    return [result_cls.from_dict(descElm) for descElm in resp.json()]
  else:
    raise to_exception(resp)

    
def to_exception(resp:requests.Response) -> MDTException:
  json_obj = resp.json()

  if 'messages' in json_obj:
    message = json_obj['messages'][0]
    return RemoteError(message['text'])
  elif 'code' in json_obj:
    code = json_obj['code']
    if code == 'java.lang.IllegalArgumentException':
      raise RemoteError(json_obj['message'])
    elif code == 'utils.InternalException':
      return RemoteError(json_obj['message'])
    elif code == 'java.lang.NullPointerException' \
        or code == 'java.lang.UnsupportedOperationException':
      raise RemoteError(f"code={json_obj['code']}, message={json_obj['message']}")
    elif code == 'org.springframework.web.servlet.resource.NoResourceFoundException':
      raise RemoteError(json_obj['text'])
    elif code == 'org.springframework.web.HttpRequestMethodNotSupportedException':
      raise RemoteError(json_obj['text'])
    
    elif code == 'mdt.model.ResourceNotFoundException':
      from .exceptions import ResourceNotFoundError
      raise ResourceNotFoundError(json_obj['message'])
    
    paths = code.split('.')
    
    from importlib import import_module
    moduleName = '.'.join(paths[:-1])
    module = import_module(moduleName)
    exception_cls = getattr(module, paths[-1])
    return exception_cls(json_obj['text'])
  else:
    return RemoteError(resp.text)



from __future__ import annotations

from typing import Any, cast, Optional, Mapping

from collections.abc import Iterator, Collection
from dataclasses import dataclass
from dataclass_wizard import JSONWizard
import datetime
from basyx.aas import model

from mdtpy.basyx.serde import from_json

from .submodel import SubmodelService
from .descriptor import MDTOperationDescriptor, MDTSubmodelDescriptor, ArgumentDescriptor
from .reference import ElementReference, DefaultElementReference, ElementReferenceDict
from .value import update_element_with_value, ElementValueType, get_value, ElementValueDict, update_value_dict, ElementJsonValueType
from .aas_misc import OperationVariable
from .exceptions import OperationError


def get_argument_value(arg:ElementReference|ElementValueType|model.SubmodelElement) -> Optional[ElementValueType]:
      if isinstance(arg, ElementReference):
        return arg.read_value()
      elif isinstance(arg, model.SubmodelElement):
        return get_value(arg)
      else:
        return arg


class MDTOperationService:
  def __init__(self, submodel_svc:SubmodelService, operation_path:str) -> None:
    self.__submodel_svc = submodel_svc
    self.__operation_path = operation_path
    operation = cast(model.Operation, submodel_svc.submodel_elements[operation_path])
    self.in_variables = [ OperationVariable(value=var) for var in operation.input_variable ]
    self.inout_variables = [ OperationVariable(value=var) for var in operation.in_output_variable ]
    self.out_variables = [ OperationVariable(value=var) for var in operation.output_variable ]

  def invoke(self, **kwargs: Any) -> ElementValueDict:
    for var in self.in_variables:
      var_id = str(var.value.id_short)
      if var_id in kwargs:
        value = get_argument_value(kwargs[var_id])
        update_element_with_value(var.value, value)
    for var in self.inout_variables:
      var_id = str(var.value.id_short)
      if var_id in kwargs:
        value = get_argument_value(kwargs[var_id])
        update_element_with_value(var.value, value)

    result = self.__submodel_svc.invoke_operation_sync(self.__operation_path, 
                                                      self.in_variables,
                                                      self.inout_variables,
                                                      timeout=datetime.timedelta(days=7))
    if result.success:
      output_values:ElementValueDict = {}
      if result.output_arguments: 
        for op_var in result.output_arguments:
          output_values[str(op_var.value.id_short)] = get_value(op_var.value)
      if result.inoutput_arguments: 
        for op_var in result.inoutput_arguments:
          output_values[str(op_var.value.id_short)] = get_value(op_var.value)
      return output_values
    else:
      if result.messages:
        raise OperationError(f'Operation {self.__operation_path} failed: {result.messages}')
      else:
        raise OperationError(f'Operation {self.__operation_path} failed')


class Argument(DefaultElementReference):
  def __init__(self, op_submodel_svc:OperationSubmodelService, desc:ArgumentDescriptor) -> None:
    super().__init__(ref_string=desc.reference,
                     endpoint=f'{op_submodel_svc.service_endpoint}/submodel-elements/{desc.id_short_path}')
    self.descriptor = desc

  @property
  def id(self) -> str:
    """
    인자 식별자를 반환한다.

    Returns:
      str: 인자 식별자.
    """
    return self.descriptor.id


class ArgumentList(ElementReferenceDict[Argument]):
  def __init__(self, op_submodel_svc:OperationSubmodelService, arg_desc_list:list[ArgumentDescriptor]):
    super().__init__({ desc.id:Argument(op_submodel_svc, desc) for desc in arg_desc_list })
    self.__arg_desc_list = arg_desc_list

  def __getitem__(self, key: str|int) -> Argument:
    if isinstance(key, int):
      return super().__getitem__(self.__arg_desc_list[key].id)
    else:
      return super().__getitem__(key)


class OperationSubmodelService(SubmodelService):
  def __init__(self, instance_id:str, sm_desc:MDTSubmodelDescriptor, op_desc:MDTOperationDescriptor) -> None:
    super().__init__(instance_id, sm_desc)
    self.__op_desc = op_desc
    self.input_arguments = ArgumentList(self, op_desc.input_arguments)
    self.output_arguments = ArgumentList(self, op_desc.output_arguments)
    self.op = MDTOperationService(self, 'Operation')

  def operation_descriptor(self) -> MDTOperationDescriptor:
    return self.__op_desc

  def invoke(self, **kwargs: Any) -> ElementValueDict:
    """
    연산을 호출한다.
    
    Args:
      kwargs: Any: 연산 인자 값을 전달하는 keyword 기반 인자.
        - 연산 식별자를 키로하는 dictionary.
        - 연산 인자 값은 전달할 값을 지정하던 ElementReference 혹은 ElementValueType 형태로 지정한다.
    
    Returns:
      ArgumentList: 연산 출력 인자 목록.
    """
    input_arg_values = self.input_arguments.read_value()
    update_value_dict(input_arg_values, kwargs)

    # 연산을 호출한다.
    result = self.op.invoke(**input_arg_values)

    # 결과 값 중에서 ElementReference 형태로 출력 인자로 제공된 경우에는
    # 해당 ElementReference 객체의 값을 갱신한다
    for arg_id, arg_value in result.items():
      if arg_id in kwargs and isinstance(kwargs[arg_id], ElementReference) \
          and arg_id in self.output_arguments:
        kwargs[arg_id].update_value(arg_value)

    return result




from __future__ import annotations

from typing import cast, Optional, Mapping, Iterator

from .descriptor import MDTParameterDescriptor
from .reference import DefaultElementReference


class MDTParameter(DefaultElementReference):
  def __init__(self, descriptor: MDTParameterDescriptor):
    super().__init__(ref_string=descriptor.reference, endpoint=cast(str, descriptor.endpoint)) 
    self.descriptor = descriptor

  @property
  def id(self) -> str:
    """
    파라미터 식별자를 반환한다.

    Returns:
      str: 파라미터 식별자.
    """
    return self.descriptor.id

  @property
  def name(self) -> Optional[str]:
    """
    파라미터 이름을 반환한다.

    Returns:
      Optional[str]: 파라미터 이름.
    """
    return self.descriptor.name


class MDTParameterCollection(Mapping[str, MDTParameter]):
  def __init__(self, parameters: list[MDTParameter]):
    self.__param_dict = { param.id:param for param in parameters }

  def __len__(self) -> int:
    return len(self.__param_dict)

  def __iter__(self) -> Iterator[str]:
    return iter(self.__param_dict.keys())

  def __contains__(self, key: str) -> bool:
    return key in self.__param_dict

  def __getitem__(self, key: str) -> MDTParameter:
    return self.__param_dict[key]


from __future__ import annotations

from typing import Optional
import re

from dateutil.relativedelta import relativedelta


_ISO8601_DURATION_RE = re.compile(
    r"""
    ^\s*
    (?P<sign>[+-])?
    P
    (?:
        (?:(?P<years>\d+(?:[.,]\d+)?)Y)?
        (?:(?P<months>\d+(?:[.,]\d+)?)M)?
        (?:(?P<weeks>\d+(?:[.,]\d+)?)W)?
        (?:(?P<days>\d+(?:[.,]\d+)?)D)?
    )?
    (?:T
        (?:(?P<hours>\d+(?:[.,]\d+)?)H)?
        (?:(?P<minutes>\d+(?:[.,]\d+)?)M)?
        (?:(?P<seconds>\d+(?:[.,]\d+)?)S)?
    )?
    \s*$
    """,
    re.VERBOSE | re.IGNORECASE,
)


def _to_number(s: Optional[str]) -> float:
    if not s:
        return 0.0
    return float(s.replace(",", "."))


def parse_iso8601_to_relativedelta(duration: str) -> relativedelta:
    """
    Parse an ISO 8601 duration string and return a dateutil.relativedelta.relativedelta.

    Examples:
      - "P1Y2M3DT4H5M6S"
      - "PT15M"
      - "P2W"
      - "-P3DT12H"
      - "PT0.5S"
    """
    m = _ISO8601_DURATION_RE.match(duration)
    if not m:
        raise ValueError(f"Invalid ISO 8601 duration: {duration!r}")

    sign = -1 if (m.group("sign") == "-") else 1

    years = _to_number(m.group("years"))
    months = _to_number(m.group("months"))
    weeks = _to_number(m.group("weeks"))
    days = _to_number(m.group("days"))
    hours = _to_number(m.group("hours"))
    minutes = _to_number(m.group("minutes"))
    seconds = _to_number(m.group("seconds"))

    # ISO 8601 allows fractional values; relativedelta expects integers for Y/M
    # (fractional months/years are ambiguous). We reject fractional Y/M explicitly.
    if years and not years.is_integer():
        raise ValueError(f"Fractional years not supported for relativedelta: {duration!r}")
    if months and not months.is_integer():
        raise ValueError(f"Fractional months not supported for relativedelta: {duration!r}")

    # For weeks/days/hours/minutes/seconds, we can carry fractions down to smaller units.
    # Convert weeks to days.
    total_days = days + (weeks * 7.0)

    # Split fractional days into seconds
    day_int = int(total_days)
    day_frac = total_days - day_int

    total_seconds = seconds + (minutes * 60.0) + (hours * 3600.0) + (day_frac * 86400.0)

    sec_int = int(total_seconds)
    sec_frac = total_seconds - sec_int
    microseconds = int(round(sec_frac * 1_000_000))

    # Handle rounding overflow (e.g., 0.9999996 rounds to 1_000_000)
    if microseconds == 1_000_000:
        sec_int += 1
        microseconds = 0

    rd = relativedelta(
        years=sign * int(years),
        months=sign * int(months),
        days=sign * day_int,
        seconds=sign * sec_int,
        microseconds=sign * microseconds,
    )
    return rd


def relativedelta_to_iso8601(rd: relativedelta) -> str:
    """
    Convert dateutil.relativedelta.relativedelta to an ISO 8601 duration string.

    Examples:
      relativedelta(years=1, months=2, days=3,
                    hours=4, minutes=5, seconds=6)
      -> "P1Y2M3DT4H5M6S"
    """

    # Determine sign (ISO-8601 uses a single leading sign)
    def _sign(x: int) -> int:
        return -1 if x < 0 else 1

    signs = [
        _sign(rd.years),
        _sign(rd.months),
        _sign(rd.days),
        _sign(rd.hours),
        _sign(rd.minutes),
        _sign(rd.seconds),
        _sign(rd.microseconds),
    ]
    sign = "-" if any(s < 0 for s in signs) else ""

    # Use absolute values (ISO-8601 sign is global)
    years = abs(rd.years)
    months = abs(rd.months)
    days = abs(rd.days)

    hours = abs(rd.hours)
    minutes = abs(rd.minutes)

    seconds = abs(rd.seconds)
    microseconds = abs(rd.microseconds)

    # Combine seconds + microseconds
    if microseconds:
        seconds = seconds + microseconds / 1_000_000

    date_parts = []
    time_parts = []

    if years:
        date_parts.append(f"{years}Y")
    if months:
        date_parts.append(f"{months}M")
    if days:
        date_parts.append(f"{days}D")

    if hours:
        time_parts.append(f"{hours}H")
    if minutes:
        time_parts.append(f"{minutes}M")
    if seconds:
        # Strip trailing .0 for integers
        if float(seconds).is_integer():
            time_parts.append(f"{int(seconds)}S")
        else:
            time_parts.append(f"{seconds:.6f}".rstrip("0").rstrip(".") + "S")

    # ISO-8601 requires at least one component
    if not date_parts and not time_parts:
        return "PT0S"

    result = sign + "P"
    result += "".join(date_parts)

    if time_parts:
        result += "T" + "".join(time_parts)

    return result


from __future__ import annotations

from typing import Any, Callable, Optional, TypeVar
from dataclasses import dataclass

import base64
import requests
import json
from urllib import parse

from dataclass_wizard import JSONWizard

from .exceptions import RemoteError, MDTException, ResourceNotFoundError
from .utils import JsonSerializable

T = TypeVar('T')

@dataclass(frozen=True, slots=True)
class Message(JSONWizard):
  message_type: str
  text: str
  code: str
  timestamp: str
    
def encode_base64url(text:str) -> str:
  """
  Encodes the given string ID to a base64 string.
  """
  return base64.urlsafe_b64encode(text.encode("utf-8")).decode("ascii")


def decode_base64url(text:str) -> str:
  """
  Decodes the given base64 string to a string.
  """
  padding = '=' * (-len(text) % 4)
  return base64.urlsafe_b64decode(text + padding).decode("utf-8")

def encode_url(text:str) -> str:
  return parse.quote(text)

def decode_url(text:str) -> str:
  return parse.unquote(text)

 
def read_none_response(resp:requests.Response) -> None:
  """
  Parses a response that is expected to have no content.
  """
  if resp.status_code >= 200 and resp.status_code < 300:
    return
  else:
    raise to_exception(resp)

  
def read_response(resp:requests.Response) -> str | None:
  """
  Parses the HTTP response based on the status code and expected result class.
  """
  if resp.status_code == 204:
    return None
  elif resp.status_code >= 200 and resp.status_code < 300:
    return resp.text
  else:
    raise to_exception(resp)
  
def read_file_response(resp:requests.Response) -> tuple[str, bytes]|None:
  if resp.status_code == 204:
    return None
  elif resp.status_code >= 200 and resp.status_code < 300:
    return resp.headers['content-type'], resp.content
  else:
    raise to_exception(resp)
    

def call_get(url:str, deserializer:Optional[Callable[[str], T]]=None) -> T|str|None:
  try:
    resp = requests.get(url, verify=False)
    resp_text = read_response(resp)
    return deserializer(resp_text) if deserializer and resp_text else resp_text
  except requests.exceptions.ConnectionError as e:
    from .exceptions import MDTInstanceConnectionError
    raise MDTInstanceConnectionError(f"Failed to connect to {url}", e)

def call_put(url:str, data:str, deserializer:Optional[Callable[[str], T]]=None) -> T|str|None:
  try:
    resp = requests.put(url, data=data, verify=False)
    resp_text = read_response(resp)
    return deserializer(resp_text) if deserializer and resp_text else resp_text
  except requests.exceptions.ConnectionError as e:
    from .exceptions import MDTInstanceConnectionError
    raise MDTInstanceConnectionError(f"Failed to connect to {url}", e)

def call_post(url:str, data:str, deserializer:Optional[Callable[[str], T]]=None) -> T|str|None:
  try:
    resp = requests.post(url, data=data, headers={'Content-Type': 'application/json'}, verify=False)
    resp_text = read_response(resp)
    return deserializer(resp_text) if deserializer and resp_text else resp_text
  except requests.exceptions.ConnectionError as e:
    from .exceptions import MDTInstanceConnectionError
    raise MDTInstanceConnectionError(f"Failed to connect to {url}", e)

def call_patch(url:str, json_str:str, deserializer:Optional[Callable[[str], T]]=None) -> T|str|None:
  try:
    resp = requests.patch(url, data=json_str, headers={'Content-Type': 'application/json'}, verify=False)
    resp_text = read_response(resp)
    return deserializer(resp_text) if deserializer and resp_text else resp_text
  except requests.exceptions.ConnectionError as e:
    from .exceptions import MDTInstanceConnectionError
    raise MDTInstanceConnectionError(f"Failed to connect to {url}", e)

def call_delete(url:str) -> None:
  try:
    resp = requests.delete(url, verify=False)
    read_none_response(resp)
  except requests.exceptions.ConnectionError as e:
    from .exceptions import MDTInstanceConnectionError
    raise MDTInstanceConnectionError(f"Failed to connect to {url}", e)


# def to_exception(resp:requests.Response) -> MDTException:
#   """
#   Converts an HTTP response to an appropriate exception based on the response content.
#   """
#   json_obj = resp.json()
#   message = Message.from_dict(json_obj['messages'][0])
#   if message.text.startswith("Resource not found"):
#       details = message.text[41:-1]
#       return ResourceNotFoundError.create("ModelRef", details)
#   elif message.text.startswith('error parsing body'):
#       return InternalError('JSON parsing failed')
#   else:
#       return MDTException(message.text)

def to_exception(resp:requests.Response) -> MDTException:
  ctype = resp.headers['Content-Type']
  if ctype.startswith('text/'):
    return RemoteError(resp.text)

  json_obj = resp.json()
  if 'messages' in json_obj:
    message = json_obj['messages'][0]
    return RemoteError(message['text'])
  elif 'code' in json_obj:
    code = json_obj['code']
    if code == 'java.lang.IllegalArgumentException':
      raise RemoteError(json_obj['message'])
    elif code == 'utils.InternalException':
      return RemoteError(json_obj['message'])
    elif code == 'java.lang.NullPointerException' \
        or code == 'java.lang.UnsupportedOperationException':
      raise RemoteError(f"code={json_obj['code']}, message={json_obj['message']}")
    elif code == 'org.springframework.web.servlet.resource.NoResourceFoundException':
      raise RemoteError(json_obj['text'])
    elif code == 'org.springframework.web.HttpRequestMethodNotSupportedException':
      raise RemoteError(json_obj['text'])
    
    elif code == 'mdt.model.ResourceNotFoundException':
      raise ResourceNotFoundError(json_obj['message'])
    
    paths = code.split('.')
    
    from importlib import import_module
    moduleName = '.'.join(paths[:-1])
    module = import_module(moduleName)
    exception_cls = getattr(module, paths[-1])
    return exception_cls(json_obj['text'])
  else:
    return RemoteError(resp.text)



from __future__ import annotations

from typing import TypeVar, Iterator, Any, Mapping, Iterable

import datetime
from urllib import parse

from basyx.aas import model

from .reference import ElementReference, DefaultElementReference
from .descriptor import MDTSubmodelDescriptor
from .aas_misc import ProtocolInformation, Endpoint, OperationVariable, OperationResult, OperationRequest, OperationHandle
from .exceptions import InvalidResourceStateError, ResourceNotFoundError
from .basyx import serde as basyx_serde
from . import fa3st


class SubmodelService:
  def __init__(self, instance_id:str, sm_desc:MDTSubmodelDescriptor) -> None:
    self.__instance_id = instance_id
    self.__descriptor = sm_desc

  @property
  def instance_id(self) -> str:
    return self.__instance_id
  
  @property
  def id(self) -> str:
    return self.__descriptor.id
  
  @property
  def id_short(self) -> str:
    return self.__descriptor.id_short

  @property
  def semantic_id_str(self) -> str:
    return self.__descriptor.semantic_id

  @property
  def service_endpoint(self) -> str|None:
    return self.__descriptor.endpoint
      
  @property
  def endpoint(self) -> Endpoint:
    return Endpoint(interface="SUBMODEL",
                    protocolInformation=ProtocolInformation(href=self.service_endpoint,
                                                            endpointProtocol="HTTP",
                                                            endpointProtocolVersion="1.1"))

  def is_information_model(self) -> bool:
    return self.__descriptor.is_information_model()

  def is_data(self) -> bool:
    return self.__descriptor.is_data()

  def is_simulation(self) -> bool:
    return self.__descriptor.is_simulation()

  def is_ai(self) -> bool:
    return self.__descriptor.is_ai()

  def is_time_series(self) -> bool:
    return self.__descriptor.is_time_series()
      
  def read(self) -> model.Submodel:
    if not self.__descriptor.endpoint:
        raise InvalidResourceStateError.create("SubmodelService", f"id={self.id}", "Endpoint is not set")
    return fa3st.call_get(self.__descriptor.endpoint, deserializer=basyx_serde.from_dict) # type: ignore
  
  def write(self, submodel:model.Submodel) -> None:
    if not self.__descriptor.endpoint:
        raise InvalidResourceStateError.create("SubmodelService", f"id={self.id}", "Endpoint is not set")
    url = f"{self.__descriptor.endpoint}"
    json_str = basyx_serde.to_json(submodel)
    fa3st.call_put(url, json_str)

  @property
  def submodel_elements(self) -> SubmodelElementCollection:
    return SubmodelElementCollection(self)

  def element_reference(self, path:str) -> ElementReference:
    ref_string = f'{self.instance_id}:{self.id_short}:{path}'
    return DefaultElementReference(ref_string=ref_string, endpoint=self.submodel_element_url(path))

  def invoke_operation_sync(self, op_path:str, input_arguments:Iterable[OperationVariable],
                            input_output_arguments:Iterable[OperationVariable],
                            timeout:datetime.timedelta) -> OperationResult:
      url = self.submodel_element_url(op_path) + "/invoke"
      req = OperationRequest(input_arguments=input_arguments,
                            inoutput_arguments=input_output_arguments,
                            client_timeout_duration=timeout)
      json_str = req.to_json()
      return fa3st.call_post(url, data=json_str, deserializer=OperationResult.from_json) # type: ignore
      
  def invoke_operation_async(self, path:str, input_arguments:list[OperationVariable],
                            input_output_arguments:list[OperationVariable],
                            timeout:datetime.timedelta) -> OperationHandle:
      url = self.submodel_element_url(path) + "/invoke?async=true"
      req = OperationRequest(input_arguments=input_arguments,
                            inoutput_arguments=input_output_arguments,
                            client_timeout_duration=timeout)
      json_str = req.to_json()
      return fa3st.call_post(url, data=json_str, serde=OperationHandle.from_json) # type: ignore
  
  def get_operation_async_result(self, path:str, handle:OperationHandle) -> OperationResult:
      url = self.submodel_element_url(path) + f"/operation-results/{handle.handle_id}"
      return fa3st.call_get(url, deserializer=OperationResult) # type: ignore

  def submodel_element_url(self, path:str) -> str:
    url_prefix = f"{self.__descriptor.endpoint}/submodel-elements"
    return f'{url_prefix}/{parse.quote(path)}' if path else url_prefix


T = TypeVar('T', bound=SubmodelService)
class SubmodelServiceCollection(Mapping[str, T]):
  def __init__(self, instance, sm_desc_dict: Mapping[str, MDTSubmodelDescriptor]):
    op_desc_dict = instance.operation_descriptors
    self.services = dict[str, T]()
    for sm_desc in sm_desc_dict.values():
      if sm_desc.is_data() or sm_desc.is_information_model():
        self.services[sm_desc.id_short] = SubmodelService(instance.id, sm_desc)  # type: ignore
      elif sm_desc.is_simulation() or sm_desc.is_ai():
        if sm_desc.id_short in op_desc_dict:
          op_desc = op_desc_dict[sm_desc.id_short]
          from .operation import OperationSubmodelService
          self.services[sm_desc.id_short] = OperationSubmodelService(instance.id, sm_desc, op_desc)  # type: ignore
        else:
          raise ResourceNotFoundError.create("MDTInstance", f"Operation {instance.id}.{sm_desc.id} not found")
      elif sm_desc.is_time_series():
        from .timeseries import TimeSeriesService
        self.services[sm_desc.id_short] = TimeSeriesService(instance.id, sm_desc)  # type: ignore

  def __bool__(self) -> bool:
    return len(self) > 0

  def __len__(self) -> int:
    return len(self.services)

  def __getitem__(self, id_short: str) -> T:
    if id_short not in self.services:
      raise ResourceNotFoundError.create("SubmodelService", f"idShort={id_short}")
    return self.services[id_short]

  def __setitem__(self, id_short: str, value) -> None:
    raise NotImplementedError('SubmodelServiceCollection does not support set operation')

  def __iter__(self) -> Iterator[str]:
    return iter(self.services.keys())

  def __contains__(self, key: str) -> bool:
    return key in self.services

  def get_by_id(self, id: str) -> T:
    for svc in self.services.values():
      if svc.id == id:
        return svc
    raise ResourceNotFoundError.create("SubmodelService", f"id={id}")

  def find_by_semantic_id(self, semantic_id: str) -> list[T]:
    return [svc for svc in self.services.values() if svc.semantic_id_str == semantic_id]


T = TypeVar('T', bound=SubmodelService)
class SubmodelServiceCollection2(Mapping[str, T]):
  def __init__(self, services: list[T]):
    self.__services = services

  def __bool__(self) -> bool:
    return len(self) > 0

  def __len__(self) -> int:
    return len(self.__services)

  def __getitem__(self, id_short: str) -> T:
    for svc in self.__services:
      if svc.id_short == id_short:
        return svc
    raise ResourceNotFoundError.create("SubmodelService", f"idShort={id_short}")

  def __setitem__(self, id_short: str, value) -> None:
    raise NotImplementedError('SubmodelServiceCollection does not support set operation')

  def __iter__(self) -> Iterator[T]:
    return iter(self.__services)

  def __contains__(self, key: str) -> bool:
    return any(svc.id == key for svc in self.__services)

  def get_by_id(self, id: str) -> T:
    for svc in self.__services:
      if svc.id == id:
        return svc
    raise ResourceNotFoundError.create("SubmodelService", f"id={id}")

  def find_by_semantic_id(self, semantic_id: str) -> list[T]:
    return [svc for svc in self.__services if svc.semantic_id_str == semantic_id]


class SubmodelElementCollection(Mapping[str, model.SubmodelElement]):
  def __init__(self, submodel_svc: SubmodelService):
    self.__submodel_svc = submodel_svc

  def element_reference(self, path: str) -> DefaultElementReference:
    ref_string = f'{self.__submodel_svc.instance_id}:{self.__submodel_svc.id_short}:{path}'
    return DefaultElementReference(ref_string=ref_string, endpoint=self.__submodel_svc.submodel_element_url(path))

  def __iter__(self) -> Iterator[str]:
    return iter(self.element_reference('').pathes())

  def __len__(self) -> int:
    return len(self.element_reference('').pathes())

  def __getitem__(self, path: str) -> model.SubmodelElement:
    return self.element_reference(path).read()

  def __setitem__(self, path: str, sme: model.SubmodelElement) -> None:
    try:
      self.element_reference(path).write(sme)
    except ResourceNotFoundError :
      self.element_reference(path).add(sme)

  def __delitem__(self, path: str) -> None:
    self.element_reference(path).remove()

  def __contains__(self, path: str) -> bool:
    return self.element_reference(path).exists()

  def get_value(self, path:str) -> Any:
    return self.element_reference(path).read_value()

  def update_value(self, path:str, value:Any) -> None:
    self.element_reference(path).update_value(value)

  def get_attachment(self, path: str) -> bytes|None:
    return self.element_reference(path).get_attachment()


from __future__ import annotations

from typing import Any
import json

from basyx.aas.adapter.json.json_deserialization import AASFromJsonDecoder
from basyx.aas.adapter.json.json_serialization import AASToJsonEncoder


def from_json(json_str:str) -> Any:
  return json.loads(json_str, cls=AASFromJsonDecoder)
  
def from_dict(data:dict) -> Any:
  json_str = json.dumps(data)
  return json.loads(json_str, cls=AASFromJsonDecoder)

def to_json(obj:Any) -> str:
  return json.dumps(obj, cls=AASToJsonEncoder)


from __future__ import annotations

from typing import cast, Iterator, Generator, Any, Optional
from abc import ABC, abstractmethod
from collections.abc import KeysView, ValuesView, ItemsView
from collections import UserDict
from dataclasses import dataclass

from datetime import datetime
from basyx.aas.model.base import Referable
from dateutil.relativedelta import relativedelta
import pandas as pd

from basyx.aas import model

import mdtpy
from .value import ElementValueType, CollectionValueType, MultiLanguagePropertyValue
from .reference import ElementReference
from . import utils


@dataclass(frozen=True, unsafe_hash=True, slots=True)
class TIMESERIES_SEMANTIC_ID:
    TIMESERIES = "https://admin-shell.io/idta/TimeSeries/1/1"
    METADATA = "https://admin-shell.io/idta/TimeSeries/Metadata/1/1"
    INTERNAL_SEGMENT = 'https://admin-shell.io/idta/TimeSeries/Segments/InternalSegment/1/1'
    LINKED_SEGMENT = 'https://admin-shell.io/idta/TimeSeries/Segments/LinkedSegment/1/1'
    EXTERNAL_SEGMENT = 'https://admin-shell.io/idta/TimeSeries/Segments/ExternalSegment/1/1'
    RECORDS = "https://admin-shell.io/idta/TimeSeries/Records/1/1"
    RECORD = "https://admin-shell.io/idta/TimeSeries/Record/1/1"


class Metadata:
  def __init__(self, metadata_value:CollectionValueType) -> None:
    self.__name = cast(MultiLanguagePropertyValue, metadata_value['Name'])
    self.__description = cast(Optional[MultiLanguagePropertyValue], metadata_value['Description'])
    self.__record = Record("rec0", cast(CollectionValueType, metadata_value['Record']))

  @property
  def name(self) -> MultiLanguagePropertyValue:
    return self.__name
  
  @property
  def description(self) -> Optional[MultiLanguagePropertyValue]:
    return self.__description
  
  @property
  def record(self) -> Record:
    return self.__record

  def __repr__(self) -> str:
    return f'Metadata(name={self.name}, fields={self.record.fields.keys()})'


class Record:
  def __init__(self, id:str, record:CollectionValueType) -> None:
    self.__id = id
    self.__fields = record

    first_field_key = next(iter(record.keys()))
    assert first_field_key is not None, f"Timestamp field is missing in record {record}"
    self.__timestamp = cast(Optional[datetime], record.get(first_field_key))

  @property
  def id(self) -> str:
    return self.__id
  
  @property
  def timestamp(self) -> Optional[datetime]:
    return self.__timestamp

  @property
  def fields(self) -> dict[str, Any|None]:
    return self.__fields
  
  def __repr__(self) -> str:
    return f'Record(id={self.id}, fields={self.fields})'


class Records:
  def __init__(self, records:CollectionValueType) -> None:
    self.__records = records
  
  def __len__(self):
    return len(self.__records)
      
  def __iter__(self) -> Generator[Record, None, None]:
    return (Record(id, cast(CollectionValueType, rec)) for id, rec in self.__records.items())


class Segment(ABC):
    def __init__(self, segment:CollectionValueType) -> None:
      self.__segment = segment

    @property
    def name(self) -> Optional[MultiLanguagePropertyValue]:
      return cast(Optional[MultiLanguagePropertyValue], self.__segment['Name'])
    
    @property
    def description(self) -> Optional[MultiLanguagePropertyValue]:
      return cast(Optional[MultiLanguagePropertyValue], self.__segment['Description'])
    
    @property
    def record_count(self) -> Optional[int]:
      return cast(Optional[int], self.__segment['RecordCount'])
    
    @property
    def start_time(self) -> Optional[datetime]:
      return cast(Optional[datetime], self.__segment['StartTime'])
    
    @property
    def end_time(self) -> Optional[datetime]:
      return cast(Optional[datetime], self.__segment['EndTime'])
    
    @property
    def duration(self) -> Optional[str]:
      return cast(Optional[str], self.__segment['Duration'])
    
    @property
    def sampling_interval(self) -> Optional[int]:
      return cast(Optional[int], self.__segment['SamplingInterval'])
    
    @property
    def sampling_rate(self) -> Optional[int]:
      return cast(Optional[int], self.__segment['SamplingRate'])
    
    @property
    def state(self) -> Optional[str]:
      return cast(Optional[str], self.__segment['State']) 
    
    @property
    def last_update(self) -> Optional[datetime]:
      return cast(Optional[datetime], self.__segment['LastUpdate'])

    @abstractmethod
    def records_as_pandas(self) -> pd.DataFrame: ...
    

class InternalSegment(Segment):
  def __init__(self, segment:CollectionValueType) -> None:
    super().__init__(segment)
    records = segment.get('Records')
    assert records is not None, f"Records is missing in InternalSegment"
    self.__records = Records(cast(CollectionValueType, records))

  @property
  def records(self) -> Records:
    return self.__records

  def records_as_pandas(self) -> pd.DataFrame:
    return pd.DataFrame([record.fields for record in self.__records])


class LinkedSegment(Segment):
  def __init__(self, segment_value:CollectionValueType) -> None:
    super().__init__(segment_value)

    endpoint = segment_value.get('Endpoint')
    assert endpoint is not None, f"Endpoint is missing in segment {segment_value}"
    self.__endpoint = cast(str, endpoint)
    query = segment_value.get('Query')
    assert query is not None, f"Query is missing in segment {segment_value}"
    self.__query = cast(str, query)

  def records_as_pandas(self) -> pd.DataFrame:
    raise NotImplementedError("LinkedSegment does not support records_as_pandas")

class ExternalSegment(Segment):
  def __init__(self, segment_smc:CollectionValueType) -> None:
    super().__init__(segment_smc)

    for sme in segment_smc.value:
      if sme.id_short == 'File':
        self.__file = cast(model.File, sme)
      elif sme.id_short == 'Blob':
        self.__blob = cast(model.Blob, sme)

  @property
  def file(self) -> model.File:
    return self.__file
  
  @property
  def blob(self) -> model.Blob:
    return self.__blob

  def records_as_pandas(self) -> pd.DataFrame:
    raise NotImplementedError("ExternalSegment does not support records_as_pandas")
        

class Segments:
    def __init__(self, segs_dict:dict[str, ElementReference]) -> None:
      def to_segment(seg_ref:ElementReference) -> Segment:
        assert (semantic_id := seg_ref.semantic_id) is not None
        values = seg_ref.read_value()
        match semantic_id.key[0].value:
          case TIMESERIES_SEMANTIC_ID.INTERNAL_SEGMENT:
            return InternalSegment(values)
          case TIMESERIES_SEMANTIC_ID.LINKED_SEGMENT:
            return LinkedSegment(values)
          case TIMESERIES_SEMANTIC_ID.EXTERNAL_SEGMENT:
            return ExternalSegment(values)
          case _:
            raise ValueError(f"Unknown segment type: {seg_ref.id_short}")

      self.__segments = { str(seg_ref.id_short):to_segment(seg_ref) for seg_name, seg_ref in segs_dict.items() }
    
    def __len__(self) -> int:
        return len(self.__segments)
    
    def __iter__(self) -> Iterator[Segment]:
        return iter(self.__segments.values())
    
    def __getitem__(self, key:str) -> Segment:
        return self.__segments[key]
    
    def __contains__(self, key:str) -> bool:
        return key in self.__segments

    def keys(self) -> KeysView[str]:
        return self.__segments.keys()
    
    def values(self) -> ValuesView[Segment]:
        return self.__segments.values()
    
    def items(self) -> ItemsView[str, Segment]:
        return self.__segments.items()
    
    def __repr__(self) -> str:
        return f'Segments({self.__segments})'


class TimeSeries:
  def __init__(self, metadata:Metadata, segments:Segments) -> None:
    self.__metadata = metadata
    self.__segments = segments

  @property
  def metadata(self) -> Metadata:
    return self.__metadata
  
  @property
  def segments(self) -> Segments:
    return self.__segments


from .descriptor import MDTSubmodelDescriptor
from .submodel import SubmodelService
class TimeSeriesService(SubmodelService):
  def __init__(self, instance_id:str, sm_desc:MDTSubmodelDescriptor) -> None:
    super().__init__(instance_id, sm_desc)

  def timeseries(self) -> TimeSeries:
    def to_segment_name(path:str) -> Optional[str]:
      path_entry_list = path.split('.')
      if len(path_entry_list) == 2:
        return path_entry_list[1]
      return None

    metadata_value = self.element_reference('Metadata').read_value()
    metadata = Metadata(metadata_value)

    segs_ref = self.element_reference('Segments')
    segs_value = segs_ref.read_value()
    
    segs_dict = { path2[1]:'.'.join(path2) for path2 in ( path.split('.') for path in segs_ref.pathes() ) if len(path2) == 2}
    segs_dict = { seg_name:self.element_reference(seg_ref_str) for seg_name, seg_ref_str in segs_dict.items() }
    
    segments = Segments(segs_dict)
    return TimeSeries(metadata, segments)


from __future__ import annotations

from typing import Any, Optional
from abc import ABC, abstractmethod

import logging

from airflow.sdk import Variable, get_current_context
from airflow.models import TaskInstance

from ..instance import MDTInstanceManager, connect
from ..operation import OperationSubmodelService
from ..reference import ElementReference, ElementValueType
from ..value import ElementValueDict
from ..exceptions import ResourceNotFoundError


def task_output(task_id:str, argument:str) -> TaskOutputArgument:
  return TaskOutputArgument(task_id, argument)

def reference(ref_string:str) -> ElementReferenceArgument:
  return ElementReferenceArgument(ref_string)

def value(value:ElementValueType) -> ValueArgument:
  return ValueArgument(value)


class DagContext:
  def __init__(self) -> None:
    self.mdt_endpoint = Variable.get("mdt_endpoint")
    self.mdt_manager = connect(self.mdt_endpoint)

  @property
  def task_id(self) -> str:
    ctx = get_current_context()
    assert ctx is not None, "Airflow context is not available"
    return ctx['task_id']

  def get_operation_submodel(self, instance:str, submodel:str) -> OperationSubmodelService:
    return self.mdt_manager.instances[instance].operations[submodel]

  def resolve_reference(self, ref_string:str) -> ElementReference:
    return self.mdt_manager.resolve_reference(ref_string)

  def get_task_instance(self) -> TaskInstance:
    ctx = get_current_context()
    assert ctx is not None, "Airflow context is not available"
    return ctx['ti']

  def get_task_output_argument(self, task_id:str, argument:str) -> ElementValueDict:
    ti = self.get_task_instance()
    logging.info(f'taskInstance: {ti}, task_id: {task_id}, argument: {argument}')

    output_args = ti.xcom_pull(task_ids=task_id, key='task_output')
    assert output_args is not None, f"Task output for {task_id} is not found"

    logging.info(f"Fetching task output for {task_id}[{argument}] from {output_args}")
    return output_args[argument]

  def set_task_output(self, task_id:str, output:ElementValueDict) -> None:
    ti = self.get_task_instance()
    logging.info(f"Setting task output for {task_id}: {output}")
    ti.xcom_push(key='task_output', value=output)


class TaskArgument(ABC):
  @abstractmethod
  def get(self, context:DagContext) -> ElementReference|Optional[ElementValueType]: ...

class TaskOutputArgument(TaskArgument):
  def __init__(self, task_id:str, argument:str) -> None:
    self.task_id = task_id
    self.argument = argument

  def get(self, context:DagContext) -> ElementReference|Optional[ElementValueType]:
    return context.get_task_output_argument(self.task_id, self.argument)

  def __repr__(self) -> str:
    return f"task_output({self.task_id}[{self.argument}])"

class ElementReferenceArgument(TaskArgument):
  def __init__(self, ref_string:str) -> None:
    self.ref_string = ref_string

  def get(self, context:DagContext) -> ElementReference|Optional[ElementValueType]:
    return context.resolve_reference(self.ref_string)

  def __repr__(self) -> str:
    return f"reference({self.ref_string})"

class ValueArgument(TaskArgument):
  def __init__(self, value:ElementValueType) -> None:
    self.value = value

  def get(self, context:DagContext) -> ElementReference|Optional[ElementValueType]:
    return self.value

  def __repr__(self) -> str:
    return f"value({self.value})"


class OperationSubmodelInvocation:
  def __init__(self, instance: str, submodel: str, arguments: dict[str, TaskArgument] ) -> None:
    self.instance = instance
    self.submodel = submodel
    self.arguments = arguments

  def run(self) -> None:
    context = DagContext()
    op_submodel = context.get_operation_submodel(self.instance, self.submodel)
    args:dict[str, ElementReference|Optional[ElementValueType]] = {}
    for arg_id, arg in self.arguments.items():
      args[arg_id] = arg.get(context)
    out_args = op_submodel.invoke(**args)
    context.set_task_output(context.task_id, out_args)

  def __repr__(self) -> str:
    return f"OperationSubmodelInvocation(task={self.instance}:{self.submodel}, arguments={self.arguments})"
